"""
Test dinghy.helpers
"""

import datetime

import freezegun
import pytest

from dinghy.helpers import find_dict_with_key, parse_since, parse_timedelta, slugify


@pytest.mark.parametrize(
    "tds, kwargs",
    [
        ("1d", dict(days=1)),
        ("1day", dict(days=1)),
        ("1d2h3m", dict(days=1, hours=2, minutes=3)),
        (
            "6 day  7.5 hours   8 min .25 s",
            dict(days=6, hours=7.5, minutes=8, seconds=0.25),
        ),
        ("10 weeks 2minutes", dict(weeks=10, minutes=2)),
    ],
)
def test_parse_timedelta(tds, kwargs):
    assert parse_timedelta(tds) == datetime.timedelta(**kwargs)


@pytest.mark.parametrize(
    "tds",
    [
        "",
        "one",
        "123",
        "1month",
        "2 years",
    ],
)
def test_bad_parse_timedelta(tds):
    assert parse_timedelta(tds) is None


@freezegun.freeze_time("2023-06-16")
@pytest.mark.parametrize(
    "since, dtargs",
    [
        ("20230730", (2023, 7, 30)),
        ("2023-06-16T12:34:56", (2023, 6, 16, 12, 34, 56)),
        ("forever", (1980, 1, 1)),
        ("1day", (2023, 6, 15)),
        ("2 weeks", (2023, 6, 2)),
        ("1 week 1 day", (2023, 6, 8)),
    ],
)
def test_parse_since(since, dtargs):
    assert parse_since(since) == datetime.datetime(*dtargs)


@pytest.mark.parametrize(
    "d, k, res",
    [
        ({"a": 1, "b": {"k": 1}, "c": "hello"}, "k", {"k": 1}),
        (
            {"a": 1, "b": {"x": 0, "d": {"k": 1, "z": 2}}, "c": "hello"},
            "k",
            {"k": 1, "z": 2},
        ),
        ({"a": 1, "b": {"k": 1}, "c": "hello"}, "z", None),
    ],
)
def test_find_dict_with_key(d, k, res):
    assert find_dict_with_key(d, k) == res


def test_slugify():
    assert slugify("Hello, World!") == "hello-world"



"""
GraphQL helpers.
"""

import asyncio
import collections
import datetime
import itertools
import logging
import os
import pkgutil
import re
import time

import aiohttp

from .helpers import DinghyError, find_dict_with_key, json_save


logger = logging.getLogger(__name__)


def _summarize_rate_limit(response):
    """
    Create a dict of information about the current rate limit.

    Reads GitHub X-RateLimit- headers.
    """
    rate_limit_info = {
        k.rpartition("-")[-1].lower(): v
        for k, v in response.headers.items()
        if k.startswith("X-RateLimit-")
    }
    rate_limit_helpfully = {
        **rate_limit_info,
        "reset_when": time.strftime(
            "%H:%M:%S",
            time.localtime(int(rate_limit_info["reset"])),
        ),
        "when": datetime.datetime.now().strftime("%H:%M:%S"),
    }
    return rate_limit_helpfully


# GraphQL error types that could be user mistakes.
USER_FIXABLE_ERR_TYPES = {
    "INSUFFICIENT_SCOPES": "Insufficient GitHub token scope.",
}


def _raise_if_error(data):
    """
    If `data` is an error response, raise a useful exception.
    """
    if "message" in data:
        raise RuntimeError(data["message"])
    if "errors" in data:
        err = data["errors"][0]
        if user_fix_msg := USER_FIXABLE_ERR_TYPES.get(err.get("type")):
            raise DinghyError(f"{user_fix_msg} {err['message']}")
        msg = f"GraphQL error: {err['message']}"
        if "path" in err:
            msg += f" @{'.'.join(err['path'])}"
        if "locations" in err:
            loc = err["locations"][0]
            msg += f", line {loc['line']} column {loc['column']}"
        logger.debug(f"Error data: {data}")
        raise RuntimeError(msg)
    if "data" in data and data["data"] is None:
        # Another kind of failure response?
        raise ValueError("GraphQL query returned null")


def _query_synopsis(query, variables):
    """
    Create a one-line synopsis of the query, for debugging and error messages.
    """
    args = ", ".join(f"{k}: {v!r}" for k, v in variables.items())
    query_head = next(line for line in query.splitlines() if not line.startswith("#"))
    return query_head + args + ")"


class GraphqlHelper:
    """
    A helper for GraphQL, including error handling and pagination.
    """

    json_names = (f"out_{i:04}.json" for i in itertools.count())
    rate_limit_history = collections.deque(maxlen=50)

    def __init__(self, endpoint, token):
        self.endpoint = endpoint
        self.headers = {"Authorization": f"Bearer {token}"}

    @classmethod
    def save_rate_limit(cls, rate_limit):
        """Keep rate limit history."""
        cls.rate_limit_history.append(rate_limit)

    @classmethod
    def last_rate_limit(cls):
        """Get the latest rate limit info."""
        if not cls.rate_limit_history:
            return None
        return cls.rate_limit_history[-1]

    async def _raw_execute(self, query, variables=None):
        """
        Execute one GraphQL query, and return the JSON data.
        """
        jbody = {"query": query}
        if variables:
            jbody["variables"] = variables
        async with aiohttp.ClientSession(headers=self.headers) as session:
            NUM_TRIES = 200
            PAUSE = 5
            total_wait = 0
            for trynum in range(NUM_TRIES):
                async with session.post(self.endpoint, json=jbody) as response:
                    if response.status == 401:
                        raise DinghyError(
                            "Unauthorized. You need to create a GITHUB_TOKEN environment variable."
                        )
                    if response.status in {403, 502} and trynum < NUM_TRIES - 1:
                        # GitHub sometimes gives us these. 403 seems like an ad-hoc
                        # unreported rate limit.  502 seems like straight-up
                        # flakiness.  If we wait them out, it goes away.
                        logger.debug(f"Wait out a 403... {total_wait} so far.")
                        await asyncio.sleep(PAUSE)
                        total_wait += PAUSE
                        continue
                    response.raise_for_status()
                    self.save_rate_limit(_summarize_rate_limit(response))
                    return await response.json()

    async def execute(self, query, variables=None):
        """
        Execute one GraphQL query, with logging, retrying, and error handling.
        """
        logger.debug(_query_synopsis(query, variables))

        while True:
            data = await self._raw_execute(query=query, variables=variables)
            if "errors" in data:
                if data["errors"][0].get("type") == "RATE_LIMITED":
                    reset_when = self.last_rate_limit()["reset_when"]
                    logger.info(f"Waiting for rate limit to reset at {reset_when}")
                    await asyncio.sleep(
                        int(self.last_rate_limit()["reset"]) - time.time() + 10
                    )
                    continue
            break

        # $set_env.py: DINGHY_SAVE_RESPONSES - save every query response in a JSON file.
        if int(os.environ.get("DINGHY_SAVE_RESPONSES", 0)):
            json_name = next(self.json_names)
            await json_save(data, json_name)
            logger.info(f"Wrote query data: {json_name}")

        _raise_if_error(data)
        return data

    async def nodes(self, query, variables=None, donefn=None, clear_nodes=True):
        """
        Execute a GraphQL query, and follow the pagination to get all the nodes.

        Returns the last query result (for the information outside the pagination),
        and the list of all paginated nodes.
        """
        nodes = []
        variables = dict(variables)
        while True:
            data = await self.execute(query, variables)
            fetched = find_dict_with_key(data, "pageInfo")
            if fetched is None:
                raise DinghyError(
                    "Query returned no data, you may need more permissions in your token: "
                    + _query_synopsis(query, variables)
                )
            nodes.extend(fetched["nodes"])
            if not fetched["pageInfo"]["hasNextPage"]:
                break
            if donefn is not None and donefn(fetched["nodes"]):
                break
            variables["after"] = fetched["pageInfo"]["endCursor"]
        # Remove the nodes from the top-level data we return, to keep things clean.
        if clear_nodes:
            fetched["nodes"] = []
        else:
            fetched["nodes"] = nodes
        return data, nodes


# $set_env.py: DINGHY_FAKE_PAGE - smaller page size to force pagination
FAKE_PAGE = int(os.environ.get("DINGHY_FAKE_PAGE", 0))


def build_query(gql_filename):
    """Read a GraphQL file, and complete it with requested fragments."""
    filenames = [gql_filename]
    query = []

    seen_filenames = set()
    while filenames:
        next_filenames = []
        for filename in filenames:
            gtext = pkgutil.get_data("dinghy", f"graphql/{filename}").decode("utf-8")
            query.append(gtext)

            for match in re.finditer(r"#\s*fragment: ([.\w]+)", gtext):
                frag_name = match[1]
                if frag_name not in seen_filenames:
                    next_filenames.append(frag_name)
                    seen_filenames.add(frag_name)
        filenames = next_filenames

    full_query = "\n".join(query)
    if FAKE_PAGE:
        full_query = full_query.replace("first: 100", f"first: {FAKE_PAGE}")
    return full_query



"""
Utilities for working with Jina2 templates.
"""

import colorsys
import datetime
from pathlib import Path

import aiofiles
import emoji
import jinja2

from dinghy.helpers import slugify


def datetime_format(value, fmt="%m-%d %H:%M"):
    """Format a datetime or ISO datetime string, for Jinja filtering."""
    if isinstance(value, str):
        value = datetime.datetime.fromisoformat(value.replace("Z", "+00:00"))
    return value.strftime(fmt)


def label_color_css(bg_color):
    """Create CSS for a label color."""
    r, g, b = [int(bg_color[i : i + 2], 16) / 255 for i in [0, 2, 4]]
    h, l, s = colorsys.rgb_to_hls(r, g, b)
    return "".join(
        f"--label-{ltr}:{int(val * fac)};"
        for ltr, val, fac in zip(
            "rgbhsl", [r, g, b, h, s, l], [255, 255, 255, 360, 100, 100]
        )
    )


def render_jinja(template_filename, **variables):
    """Render a template file, with variables."""
    jenv = jinja2.Environment(
        loader=jinja2.FileSystemLoader(
            [
                Path("."),
                Path(__file__).parent / "templates",
            ]
        ),
        autoescape=True,
    )
    jenv.filters["datetime"] = datetime_format
    jenv.filters["label_color_css"] = label_color_css
    jenv.filters["slugify"] = slugify
    template = jenv.get_template(template_filename)
    html = template.render(**variables)
    return html


async def render_jinja_to_file(template_filename, output_file, **variables):
    """Render a template file with variables, and write it to a file."""
    text = render_jinja(template_filename, **variables)
    text = emoji.emojize(text, language="alias")
    async with aiofiles.open(output_file, "w", encoding="utf-8") as out:
        await out.write(text)



"""
Summarize issue activity in GitHub repos and projects.
"""

import asyncio
import datetime
import json
import logging
import operator
import os
import re
import urllib.parse

import yaml
from glom import glom

from . import __version__
from .graphql_helpers import build_query, GraphqlHelper
from .helpers import DinghyError, json_save, parse_since
from .jinja_helpers import render_jinja_to_file


logger = logging.getLogger(__name__)

GITHUB_URL_MAP = []


def github_route(url_pattern):
    """A decorator to associate a GitHub URL path regex with a Digester.get_ method.

    The regexes will be tried in the order the decorator is used in the class,
    so be careful if a path could match multiple patterns.
    """

    def _dec(func):
        GITHUB_URL_MAP.append((url_pattern, func.__name__))
        return func

    return _dec


def dd(name: str) -> str:
    """Create an key for Dinghy-specfic data. (dd == Dinghy Data)

    Dinghy adds extra data to GitHub responses. All of the keys for this data
    are prefixed, created with this function.
    """
    return f"dinghy_{name}"


DD_children = dd("children")


class Digester:
    """
    Use GitHub GraphQL to get data about recent changes.
    """

    def __init__(self, since, options):
        self.since = since.strftime("%Y-%m-%dT%H:%M:%S")
        self.ignore_users = options.get("ignore_users", [])
        self.user_types = {"User"}
        if options.get("include_bots", False):
            self.user_types.add("Bot")
        self.api_root = options.get("api_root")
        self.github = "github.com"
        self.gql = None

    def prepare(self):
        """Create the network helpers we need."""
        token = os.environ.get("GITHUB_TOKEN", "")
        api_root = self.api_root or f"https://api.{self.github}/graphql"
        self.gql = GraphqlHelper(api_root, token)

    @github_route(r"/orgs/(?P<org>[^/]+)/projects/(?P<number>\d+)/?")
    async def get_org_project_entries(self, org, number, home_repo="", title=None):
        """
        Get entries from a organization project.

        Args:
            org (str): the organization owner of the repo.
            number (int|str): the project number.
            home_repo (str): the owner/name of a repo that most entries are in.
        """
        project, project_data = await self.gql.nodes(
            query=build_query("org_project_entries.graphql"),
            variables=dict(org=org, projectNumber=int(number)),
        )
        entries = [content for data in project_data if (content := data["content"])]
        entries = await self._process_entries(entries)
        for entry in entries:
            entry["other_repo"] = entry["repository"]["nameWithOwner"] != home_repo
            if DD_children not in entry:
                entry[DD_children] = entry["comments"]["nodes"]
        project = glom(project, "data.organization.project")
        container = {
            "url": project["url"],
            "container_kind": "project",
            "title": title or project["title"],
            "kind": "items",
            "entries": entries,
        }
        return container

    async def get_search_results(self, query, title=None):
        """
        Get issues or pull requests returned by a search query.
        """
        query += f" updated:>{self.since}"
        _, entries = await self.gql.nodes(
            query=build_query("search_entries.graphql"),
            variables=dict(query=query),
        )
        entries = await self._process_entries(entries)
        for entry in entries:
            entry["other_repo"] = True
        url_q = urllib.parse.quote_plus(query)
        if "is:pr" in query:
            kind = "pull requests"
        elif "is:issue" in query:
            kind = "issues"
        else:
            kind = "items"
        container = {
            "url": f"https://{self.github}/search?q={url_q}&type=issues",
            "container_kind": "search",
            "title": title or query,
            "kind": kind,
            "entries": entries,
        }
        return container

    @github_route(r"/(?P<owner>[^/]+)/(?P<name>[^/]+)/?")
    async def get_repo_entries(self, owner, name, title=None):
        """
        Get issues, pull requests, and releases from a repo.
        """
        issue_container, pr_container, release_container = await asyncio.gather(
            self.get_repo_issues(owner, name, title=title),
            self.get_repo_pull_requests(owner, name),
            self.get_repo_releases(owner, name),
        )
        entries = (
            issue_container["entries"]
            + pr_container["entries"]
            + release_container["entries"]
        )
        entries = self._trim_unwanted(entries)
        container = {
            **issue_container,
            "kind": "issues, pull requests, and releases",
            "entries": entries,
        }
        return container

    @github_route(r"/(?P<owner>[^/]+)/(?P<name>[^/]+)/issues/?")
    async def get_repo_issues(self, owner, name, title=None):
        """
        Get issues from a repo updated since a date, with comments since that date.

        Args:
            owner (str): the owner of the repo.
            name (str): the name of the repo.
        """
        repo, issues = await self.gql.nodes(
            query=build_query("repo_issues.graphql"),
            variables=dict(owner=owner, name=name, since=self.since),
        )
        issues = await self._process_entries(issues)
        repo = glom(repo, "data.repository")
        container = {
            "url": repo["url"],
            "container_kind": "repo",
            "title": title or repo["nameWithOwner"],
            "kind": "issues",
            "entries": issues,
        }
        return container

    @github_route(r"/(?P<owner>[^/]+)/(?P<name>[^/]+)/releases/?")
    async def get_repo_releases(self, owner, name, title=None):
        """
        Get releases from a repo updated since a date

        Args:
            owner (str): the owner of the repo.
            name (str): the name of the repo.
        """
        repo, releases = await self.gql.nodes(
            query=build_query("repo_releases.graphql"),
            variables=dict(owner=owner, name=name, since=self.since),
        )
        releases = await self._process_entries(releases)
        repo = glom(repo, "data.repository")
        container = {
            "url": repo["url"],
            "container_kind": "repo",
            "title": title or repo["nameWithOwner"],
            "kind": "releases",
            "entries": releases,
        }
        return container

    @github_route(r"/(?P<owner>[^/]+)/(?P<name>[^/]+)/pulls/?")
    async def get_repo_pull_requests(self, owner, name, title=None):
        """
        Get pull requests from a repo updated since a date, with comments since that date.

        Args:
            owner (str): the owner of the repo.
            name (str): the name of the repo.
        """
        repo, pulls = await self.gql.nodes(
            query=build_query("repo_pull_requests.graphql"),
            variables=dict(owner=owner, name=name),
            donefn=(lambda nodes: nodes[-1]["updatedAt"] < self.since),
        )
        pulls = await self._process_entries(pulls)

        repo = glom(repo, "data.repository")
        container = {
            "url": repo["url"],
            "container_kind": "repo",
            "title": title or repo["nameWithOwner"],
            "kind": "pull_requests",
            "entries": pulls,
        }
        return container

    def method_from_url(self, url):
        """
        Dispatch to a get_* method from a GitHub URL.

        Args:
            url (str): A GitHub URL

        Returns:
            A method, and a dict of **kwargs.
        """
        parsed = urllib.parse.urlparse(url)
        self.github = parsed.netloc
        for rx, fn_name in GITHUB_URL_MAP:
            if match_url := re.fullmatch(rx, parsed.path):
                return getattr(self, fn_name), match_url.groupdict()

        raise DinghyError(f"Can't understand URL {url!r}")

    async def get_more(
        self, container, graphql, id
    ):  # pylint: disable=redefined-builtin
        """
        If the `container` isn't full yet, get all the nodes.
        """
        if container["totalCount"] > len(container["nodes"]):
            _, all_nodes = await self.gql.nodes(
                query=build_query(graphql),
                variables=dict(id=id),
            )
            container["nodes"] = all_nodes

    def _node_is_interesting(self, node):
        """
        Is a node interesting to show? It has to be new enough, by a real user,
        and not by someone we want to ignore.
        """
        return (
            node["updatedAt"] > self.since
            and node["author"]["__typename"] in self.user_types
            and node["author"]["login"] not in self.ignore_users
        )

    def _trim_unwanted(self, nodes):
        """
        Trim a list to keep only activity since `self.since`, and only by real
        users.

        The returned list is also sorted by updatedAt date.
        """
        nodes = (n for n in nodes if self._node_is_interesting(n))
        nodes = sorted(nodes, key=operator.itemgetter("updatedAt"))
        return nodes

    def _fix_ghosts(self, obj):
        """
        GitHub has a @ghost account for deleted users. That shows up in our
        data as None.  Fix those to have data we can use.
        """
        if isinstance(obj, list):
            for elt in obj:
                self._fix_ghosts(elt)
        elif isinstance(obj, dict):
            for key in obj:
                if key == "author":
                    if obj["author"] is None:
                        obj["author"] = {
                            "__typename": "User",
                            "login": "ghost",
                        }
                else:
                    self._fix_ghosts(obj[key])

    async def _process_entries(self, entries):
        """
        Process entries after they've been retrieved.

        Keep only things updated since our date, and sort them.
        """
        # $set_env.py: DINGHY_SAVE_ENTRIES - save each entry in its own JSON file.
        if int(os.environ.get("DINGHY_SAVE_ENTRIES", 0)):
            for entry in entries:
                try:
                    kind = entry["__typename"].lower()
                    num = entry["number"]
                except KeyError:
                    pass
                else:
                    await json_save(entry, f"save_{kind}_{num}.json")

        self._fix_ghosts(entries)

        entries = self._trim_unwanted(entries)
        entries = await asyncio.gather(*map(self._process_entry, entries))
        return entries

    async def _process_entry(self, entry):
        """
        Apply entry-specific processing to an entry.
        """
        if entry["__typename"] == "Issue":
            await self._process_issue(entry)
        elif entry["__typename"] == "PullRequest":
            await self._process_pull_request(entry)
        self._add_reasons(entry)
        return entry

    async def _process_issue(self, issue):
        """
        Add more comments to an issue.

        We can't paginate the comments on issues while paginating issues, so
        this method gets the rest of the comments.

        Args:
            issue (dict): the issue to populate.
        """
        await self.get_more(issue["comments"], "issue_comments.graphql", issue["id"])
        issue[DD_children] = self._trim_unwanted(issue["comments"]["nodes"])

    async def _process_pull_request(self, pull):
        """
        Do extra work to make a pull request right for reporting.
        """
        # Pull requests have complex trees of data, with comments in
        # multiple places, and duplications.  Reviews can also be finished
        # with no comment, but we want them to appear in the digest.
        #
        # Pull requests have:
        #   comments:
        #       Standalone comments that should always be included
        #   reviews:
        #       Each is a review by a person, who can add comments all over the
        #       pull request, including in different threads.
        #   reviewThreads:
        #       Each is a sequence of comments that follow one another.
        #

        # Pull all the data from paginated components.
        await asyncio.gather(
            self.get_more(pull["comments"], "pr_comments.graphql", pull["id"]),
            self.get_more(pull["reviews"], "pr_reviews.graphql", pull["id"]),
            self.get_more(
                pull["reviewThreads"], "pr_reviewthreads.graphql", pull["id"]
            ),
        )

        coros = []
        coros.extend(
            self.get_more(r["comments"], "review_comments.graphql", r["id"])
            for r in pull["reviews"]["nodes"]
        )
        coros.extend(
            self.get_more(rt["comments"], "reviewthread_comments.graphql", rt["id"])
            for rt in pull["reviewThreads"]["nodes"]
        )
        await asyncio.gather(*coros)

        children = {}
        reviews = {}

        # Make a map of the reviews.
        for rev in pull["reviews"]["nodes"]:
            rev[dd("review_state")] = rev["state"]
            reviews[rev["id"]] = rev

        # For each thread, attach the thread as a child of the review.  Each
        # comment in the thread can be from a different review (as people
        # respond to each other).  The whole thread will be attached to the
        # review for the first comment.  Make comments 2-N as children of
        # comment 1.
        for thread in pull["reviewThreads"]["nodes"]:
            com0 = thread["comments"]["nodes"][0]
            com0[DD_children] = thread["comments"]["nodes"][1:]
            com0["isResolved"] = thread["isResolved"]
            if com0["pullRequestReview"]:
                rev_id = com0["pullRequestReview"]["id"]
                review_comments = reviews[rev_id].setdefault(DD_children, [])
                review_comments.append(com0)

        # For each review, show it if it has a body, or if it has children, or
        # if it's not just "COMMENTED".
        for rev in reviews.values():
            if rev["bodyText"] or rev.get(DD_children) or rev["state"] != "COMMENTED":
                com = children.setdefault(rev["id"], dict(rev))
                com[dd("review_state")] = rev["state"]

                if not rev["bodyText"] and len(rev.get(DD_children, ())) == 1:
                    # A review with just one comment and no body: the comment should
                    # go where the review would have been.
                    com = rev[DD_children][0]
                    com[dd("review_state")] = rev[dd("review_state")]
                    children[rev["id"]] = com

        # Comments are simple: they all get shown.
        for com in pull["comments"]["nodes"]:
            children[com["id"]] = com

        # Examine all the resulting threads (children). Keep a thread if it has
        # any comments newer than our since date.  Mark older comments as old.
        kids, _ = self._trim_unwanted_tree(children.values())
        pull[DD_children] = kids

    def _trim_unwanted_tree(self, nodes):
        """
        Trim a nested list to indicate activity since `self.since`.  A thread
        will be kept if any of its children is newer than since.  Items older
        than that will get ["boring"]=True, and shown grayed in the output.
        """
        keep = []
        any_interesting_total = False
        for node in nodes:
            if self._node_is_interesting(node):
                any_interesting = True
            else:
                any_interesting = False
                node[dd("boring")] = True
            kids, any_interesting_kids = self._trim_unwanted_tree(
                node.get(DD_children, ())
            )
            if any_interesting or any_interesting_kids:
                node[DD_children] = kids
                keep.append(node)
                any_interesting_total = True
        keep = sorted(keep, key=operator.itemgetter("updatedAt"))
        return keep, any_interesting_total

    def _add_reasons(self, entry):
        """
        Populate an entry with the reasons it's been included.

        Args:
            entry (dict): the issue or pull request data.

        """
        # write "reasonCreated" based on "createdAt", etc.
        for slug in ["Created", "Closed", "Merged"]:
            at = slug.lower() + "At"
            entry[dd(f"reason{slug}")] = bool(entry.get(at) and entry[at] > self.since)


def coro_from_item(digester, item):
    """
    Parse a single config item, and make a digester coro for it.
    """
    url = None
    more_kwargs = {}
    if isinstance(item, str):
        url = item
    elif "url" in item:
        more_kwargs = dict(item)
        url = more_kwargs.pop("url")

    if url:
        fn, kwargs = digester.method_from_url(url)
    else:
        if "search" in item:
            kwargs = dict(item)
            kwargs["query"] = kwargs.pop("search")
            fn = digester.get_search_results
        else:
            raise DinghyError(f"Don't understand item: {item!r}")

    try:
        coro = fn(**kwargs, **more_kwargs)
    except TypeError as type_err:
        raise DinghyError(f"Problem with config item: {item}: {type_err}") from None

    return coro


async def make_digest(items, since=None, digest="digest.html", **options):
    """
    Make a single digest.

    Args:
        since (optional str): a duration spec ("2 day", "3d6h", etc). Default: 1 week.
        items (list[str|dict]): a list of YAML objects or GitHub URLs to collect entries from.
        digest (str): the HTML file name to write.

    """
    if since is None:
        since = "1 week"
    show_date = since != "forever"
    since_date = parse_since(since)
    digester = Digester(since=since_date, options=options)

    coros = []
    for item in items:
        try:
            coros.append(coro_from_item(digester, item))
        except:
            for coro in coros:
                coro.close()
            raise

    digester.prepare()
    results = await asyncio.gather(*coros)

    # $set_env.py: DINGHY_SAVE_RESULT - save digest data in a JSON file.
    if int(os.environ.get("DINGHY_SAVE_RESULT", 0)):
        json_name = digest.replace(".html", ".json")
        await json_save(results, json_name)
        logger.info(f"Wrote results data: {json_name}")

    await render_jinja_to_file(
        options.get("template", "digest.html.j2"),
        digest,
        results=results,
        since=since_date if show_date else None,
        now=datetime.datetime.now(),
        __version__=__version__,
        title=options.get("title", ""),
    )
    logger.info(f"Wrote digest: {digest}")


async def make_digests_from_config(conf_file, digests=None, since=None):
    """
    Make all the digests specified by a configuration file.

    Args:
        conf_file (str): a file path to read as a config file.
        digests (list of str): the digest names to make.
        since (str): the spec for since when.
    """
    try:
        with open(conf_file, encoding="utf-8") as cf:
            config = yaml.safe_load(cf)
    except Exception as err:
        raise DinghyError(f"Couldn't read config file {conf_file!r}: {err}") from err

    if "digests" not in config:
        raise DinghyError(f"No 'digests:' clause in config file {conf_file!r}")

    defaults = config.get("defaults", {})
    coros = []
    for spec in config["digests"]:
        args = {**defaults, **spec}
        if digests is not None and args["digest"] not in digests:
            continue
        if since is not None:
            args["since"] = since
        coros.append(make_digest(**args))
    await asyncio.gather(*coros)


def just_render(result_file):
    """Helper function to re-render stored results.

    For iterating on rendering changes without using up GitHub rate limits.

    $ python -c "import sys,dinghy.digest as dd; dd.just_render(sys.argv[1])" /tmp/lots.json

    """
    with open(result_file, encoding="utf-8") as j:
        results = json.load(j)

    asyncio.run(
        render_jinja_to_file(
            "digest.html.j2",
            result_file.replace(".json", ".html"),
            results=results,
            since=datetime.datetime.now(),
            now=datetime.datetime.now(),
            __version__=__version__,
        )
    )



"""
A module-main for running ad-hoc GitHub GraphQL queries.

After installing dinghy, run it like this:

    $ python -m dinghy.adhoc --help

"""

import json
import os
import sys

import click
import click_log

from .cli import main_run
from .graphql_helpers import GraphqlHelper


logger = click_log.basic_config("dinghy")


TYPES = {
    "int": int,
    "str": str,
}


@click.command()
@click_log.simple_verbosity_option(logger)
@click.option(
    "--nodes",
    is_flag=True,
    help="Get paginated list of nodes instead of raw result",
)
@click.argument("query_file", type=click.File("r"))
@click.argument("var", metavar="[VAR[:type]=VAL]...", nargs=-1)
def adhoc(nodes, query_file, var):
    """
    Run an ad-hoc GraphQL query.
    """
    query = query_file.read()
    variables = {}
    for v in var:
        name, val = v.split("=", 1)
        if ":" in name:
            name, type_name = name.split(":")
            val = TYPES[type_name](val)
        variables[name] = val

    token = os.environ.get("GITHUB_TOKEN", "")
    gql = GraphqlHelper("https://api.github.com/graphql", token)
    if nodes:
        data, _ = main_run(
            gql.nodes(query=query, variables=variables, clear_nodes=False)
        )
    else:
        data = main_run(gql.execute(query=query, variables=variables))
    json.dump(data, sys.stdout, indent=2)


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    adhoc()



"""
Misc helpers.
"""

import datetime
import json
import re
import sys
import unicodedata

import aiofiles


if sys.version_info < (3, 11):
    from backports.datetime_fromisoformat import MonkeyPatch
    MonkeyPatch.patch_fromisoformat()


class DinghyError(Exception):
    """An error in how Dinghy is being used."""


async def json_save(data, filename):
    """Write `data` to `filename` as JSON."""
    async with aiofiles.open(filename, "w", encoding="utf-8") as json_out:
        await json_out.write(json.dumps(data, indent=4))


def parse_timedelta(timedelta_str):
    """
    Parse a timedelta string ("2h13m") into a timedelta object.

    From https://stackoverflow.com/a/51916936/14343

    Args:
        timedelta_str (str): A string identifying a duration, like "2h13m".

    Returns:
        A datetime.timedelta object, or None if it can't be parsed.

    """
    parts = re.match(
        r"""(?x)
        ^
        ((?P<weeks>[.\d]+)w(eeks?)?)?
        ((?P<days>[.\d]+)d(ays?)?)?
        ((?P<hours>[.\d]+)h(ours?)?)?
        ((?P<minutes>[.\d]+)m(in(utes?)?)?)?
        ((?P<seconds>[.\d]+)s(ec(onds?)?)?)?
        $
        """,
        timedelta_str.replace(" ", ""),
    )
    if not timedelta_str or parts is None:
        return None
    kwargs = {name: float(val) for name, val in parts.groupdict().items() if val}
    return datetime.timedelta(**kwargs)


def parse_since(since):
    """
    Parse a since specification:

    - "forever" uses a long-ago date.
    - A time delta (like "1 week") computes that long ago.
    - A specific time (like "2023-07-30") is used as-is.

    """
    if since == "forever":
        since_date = datetime.datetime(year=1980, month=1, day=1)
    else:
        delta = parse_timedelta(since)
        if delta is not None:
            since_date = datetime.datetime.now() - delta
        else:
            try:
                since_date = datetime.datetime.fromisoformat(since)
            except ValueError:
                raise DinghyError(f"Can't parse 'since' value: {since!r}") from None
    return since_date


def find_dict_with_key(d, key):
    """Return the subdict of `d` that has `key`."""
    if key in d:
        return d
    for dd in d.values():
        if isinstance(dd, dict):
            sd = find_dict_with_key(dd, key)
            if sd is not None:
                return sd
    return None


def slugify(value, allow_unicode=False):
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert to lowercase. Also strip leading and
    trailing whitespace, dashes, and underscores.

    from django.template.defaultfilters import slugify
    """
    value = str(value).strip()
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")



"""Dinghy command-line interface."""

import asyncio
import sys

import click
import click_log

from .digest import make_digest, make_digests_from_config
from .graphql_helpers import GraphqlHelper
from .helpers import DinghyError

# Fix for https://github.com/nedbat/dinghy/issues/9
# Work around a known problem (https://github.com/python/cpython/issues/83413)
# that is fixed in 3.10.6 (https://github.com/python/cpython/pull/92904).
if sys.version_info < (3, 10, 6) and sys.platform.startswith("win"):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

logger = click_log.basic_config("dinghy")


def main_run(coro):
    """
    Run a coroutine for a Dinghy command.
    """
    try:
        return asyncio.run(coro)
    except DinghyError as err:
        logger.error(f"dinghy error: {err}")
        sys.exit(1)
    finally:
        lrl = GraphqlHelper.last_rate_limit()
        if lrl is not None:
            resource = lrl.get("resource", "general")
            logger.debug(
                f"Remaining {resource} rate limit: "
                + f"{lrl['remaining']} of {lrl['limit']}, "
                + f"next reset at {lrl['reset_when']}"
            )


@click.command()
@click_log.simple_verbosity_option(logger)
@click.version_option()
@click.option(
    "--since",
    metavar="DELTA-OR-DATE",
    help="Specify a since date. [default: 1 week]",
)
@click.argument("_input", metavar="[INPUT]", default="dinghy.yaml")
@click.argument("digests", metavar="[DIGEST ...]", nargs=-1)
def cli(since, _input, digests):
    """
    Generate HTML digests of GitHub activity.

    INPUT is a dinghy YAML configuration file (default: dinghy.yaml), or a
    GitHub repo URL.

    DIGEST(s) are the file names of digests from the configuration file to
    create.  If none are specified, all of the digests are written.

    """
    if "://" in _input:
        coro = make_digest([_input], since=since)
    else:
        coro = make_digests_from_config(_input, digests or None, since=since)

    main_run(coro)
